#------------------------------------------------------------------------------

# MIT License

# Copyright (c) 2020 Landcare Research New Zealand Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

# Python 3.5.5 | packaged by conda-forge | (default, Jul 24 2018, 01:52:17) [MSC v.1900 64 bit (AMD64)] on win32
import numpy as np # version 1.14.2
from scipy import spatial # version 1.1.0

#------------------------------------------------------------------------------

def rasterDomainCellCoordinates(C):
    """    
    Calculate coordinates of cell centres for a specified raster spatial 
    domain.

    Parameters
    ----------
    C : list
        List of values that defines the spatial extent and resolution of the
        interpolated surface.  This must be specificed in the form of        
        [xMin, yMin, nRow, nCol, cellSize] where xMin and yMin are the x and y
        coordinates of the bottom left corner of the extent, nRow and nCol give
        the number of rows and columns, and cellSize gives the size of the 
        cells in the interpolated surface.
        
    Returns
    -------
    out : array
        c x 2 array of the x and y coordinates of all cell centres.
    """
    xMin, yMin, nRow, nCol, cellSize = C
    xMax = xMin + (nCol * cellSize)
    yMax = yMin + (nRow * cellSize)
    halfCell = (cellSize / 2)
    c_x = np.linspace(xMin + halfCell, xMax - halfCell, nCol)
    c_y = np.linspace(yMax - halfCell, yMin + halfCell, nRow)
    c_xy = np.column_stack((np.tile(c_x, nRow), np.repeat(c_y, nCol)))
    return(c_xy)

def xyzToDataCells(xyz, C):
    """    
    Converts data points to data cells by moving the points to the centre of 
    the cell in which they occur.  Points falling in the same cell are combined
    into a single data cell by taking the mean z value.  The process tries to
    keep the index the same, but when data points are merged into a data cell 
    the indices will change slightly as the total number of cells will be less
    then the number of points.

    Parameters
    ----------
    xyz : array
        A n x 3 array in which each row is a data point, and each column is the
        x coordinate, y coordinate, and z value.
    C : list
        List of values that defines the spatial extent and resolution of the
        interpolated surface.  This must be specificed in the form of        
        [xMin, yMin, nRow, nCol, cellSize] where xMin and yMin are the x and y
        coordinates of the bottom left corner of the extent, nRow and nCol give
        the number of rows and columns, and cellSize gives the size of the 
        cells in the interpolated surface.
        
    Returns
    -------
    out : array
        n x 3 array of data cells.
    """
    xMin, yMin, nRow, nCol, cellSize = C
    halfCell = (cellSize / 2)
    col = np.floor((xyz[:,0] - xMin) / cellSize)
    row = nRow - np.floor((xyz[:,1] - yMin) / cellSize) - 1
    p_x = xMin + (col * cellSize) + halfCell
    p_y = yMin + ((nRow - row - 1) * cellSize) + halfCell
    P = np.column_stack((p_x, p_y, xyz[:,2]))
    # Check for duplicate cells and where present take the mean value
    cells = {}
    cellIDs = []
    for i in range(P.shape[0]):
        cellID = str(P[i,0]) + "-" + str(P[i,1])
        if cellID not in cells.keys():
            cells[cellID] = [P[i,2]]
            cellIDs.append(cellID)
        else:
            cells[cellID].append(P[i,2])
    xs = []
    ys = []
    zs = []
    # Array the data cells in the same order as the data points if possible
    for cellID in cellIDs:
        x, y = cellID.split("-")
        xs.append(float(x))
        ys.append(float(y))
        zs.append(np.mean(cells[cellID]))
    P = np.column_stack((xs, ys, zs))
    return(P)

def discreteVoronoiDiagram(P, c_xy):
    """    
    Calculates the discrete Voronoi diagram (VP) and the distance from the data
    cells to all raster cells (DPC).

    Parameters
    ----------
    P : array
        n x 3 array of data cells in which each row is a data cell, and each 
        column is the x coordinate, y coordinate, z value.
    c_xy : array
        c x 2 array of the x and y coordinates of all cell centres.
        
    Returns
    -------
    out : tuple
        Out tuple contains in order: the data cell indices for the Voronoi 
        polygons (VP), and the distance from the data cells to the raster cells 
        (DPC).
    """
    # Create a k-d tree for the data cells
    kdtreeData = spatial.cKDTree(P[:,:2])
    # Extract the distance and index to the nearest data cell
    DPC, VP = kdtreeData.query(c_xy, k=1)
    return((VP, DPC))

def distanceFromCell(fromCell, toCells):
    """    
    Calculates the Euclidean distance from a cell to other cells.

    Parameters
    ----------
    fromCell : array
        Length 2 array of x and y coordinates for a cell from which to measure 
        distance.
    toCells : array
        n x 2 array of x and y coordinates for cells to measure distance to.
        
    Returns
    -------
    out : array
        Length n array of distances.
    """
    x, y = fromCell
    d = np.hypot(toCells[:,0] - x, toCells[:,1] - y)
    return(d)

def interpolationCellVoronoi(DPC, Dc_iC):
    """    
    Determines which raster domaon cells belong to the discrete Voronoi polygon
    for an interpolation cell.

    Parameters
    ----------
    DPC : array
        The distances from the data cells (P) to the raster cells (C).
    Dc_iC : array
        The distances from the interpolation cell (c_i) to the raster cells (C).
        
    Returns
    -------
    out : array
        Raster cells belonging to the Voronoi polygon are coded 1.
    """   
    
    Vc_i = np.ones((DPC.shape))
    np.place(Vc_i, Dc_iC > DPC, np.nan)
    return(Vc_i)

def crossValidateRateOfError(P, c_xy):
    """    
    Conduct cross-validation to calculate the rate of absolute error for each
    data cell.

    Parameters
    ----------
    P : array
        n x 3 array of data cells in which each row is a data cell, and each 
        column is the x coordinate, y coordinate, z value.
    c_xy : array
        c x 2 array of the x and y coordinates of all cell centres.
        
    Returns
    -------
    out : array
        n x 4 array of data cells in which each row is a data cell, and each 
        column is the x coordinate, y coordinate, z value, and r rate of error.
    """
    r = [] # rate of absolute error
    index = list(np.arange(P.shape[0]))
    for i in index:
        # Create indices of training and testing data
        testCV = [i]
        trainCV = index[:i] + index[i + 1:]
        # Extract training and testing data
        testData = P[testCV, 0:2]
        trainData = P[trainCV, :]
        # Cross-validate
        VP, DPC = discreteVoronoiDiagram(trainData, c_xy)
        ZP = trainData[VP, 2]
        Dc_iC = distanceFromCell(testData[0], c_xy)
        Vc_i = interpolationCellVoronoi(DPC, Dc_iC)
        zest = np.nanmean(ZP * Vc_i) # value estimate
        abse = np.abs(P[i,2] - zest) # absolute error
        p_j = distanceFromCell(testData[0], trainData)
        DP = p_j[VP]
        nnd = np.nanmean(DP * Vc_i) # natural neighbour distance
        r.append(abse / nnd)
    return(np.column_stack((P, r)))

def crossValidateMeanAbsoluteError(P, c_xy):
    """    
    Calculate the cross-validation mean absolute error for the data cells.

    Parameters
    ----------
    P : array
        n x 3 array of data cells in which each row is a data cell, and each 
        column is the x coordinate, y coordinate, z value.
    c_xy : array
        c x 2 array of the x and y coordinates of all cell centres.
        
    Returns
    -------
    out : float
        The mean absolute error.
    """
    absoluteErrors = []
    index = list(np.arange(P.shape[0]))
    for i in index:
        testCV = [i]
        trainCV = index[:i] + index[i + 1:]
        # Extract training and testing data
        testData = P[testCV, 0:2]
        trainData = P[trainCV, :]
        # Cross-validate
        VP, DPC = discreteVoronoiDiagram(trainData, c_xy)
        ZP = trainData[VP, 2]
        Dc_iC = distanceFromCell(testData[0], c_xy)
        Vc_i = interpolationCellVoronoi(DPC, Dc_iC)
        zest = np.nanmean(ZP * Vc_i) # value estimate
        abse = np.abs(P[i,2] - zest) # absolute error
        absoluteErrors.append(abse)
    return(np.mean(absoluteErrors))

def nni(xyz, C, extras=False):
    """    
    Calculate the natural neighbour interpolation field and the cross-
    validation error-distance field.

    Parameters
    ----------
    xyz : array
        A n x 3 array in which each row is a data point, and each column is the
        x coordinate, y coordinate, and z value.
    C : list
        List of values that defines the spatial extent and resolution of the
        interpolated surface.  This must be specificed in the form of        
        [xMin, yMin, nRow, nCol, cellSize] where xMin and yMin are the x and y
        coordinates of the bottom left corner of the extent, nRow and nCol give
        the number of rows and columns, and cellSize gives the size of the 
        cells in the interpolated surface.
    extras : boolean
        If extras == True then the data cell natural neighbour distance field 
        and the estimated rate of error fields are also returned.
        
    Returns
    -------
    out : Tuple
        Raster fields for the natural neighbour interpolation and the cross-
        validation error-distance, and if extras == True also the natural
        neighbour distance and estimated rate of error.
    """ 
    # Extract raster metadata
    xMin, yMin, nRow, nCol, cellSize = C    
    # Get xy coordinates for raster cells
    c_xy = rasterDomainCellCoordinates(C)    
    # Convert points to cell centres
    P = xyzToDataCells(xyz, C)
    # Conduct cross-validation to get absolute rates of error
    P = crossValidateRateOfError(P, c_xy)
    # Calculate the discrete Voronoi diagram
    VP, DPC = discreteVoronoiDiagram(P, c_xy)
    ZP = P[VP, 2] # Extract values by Voronoi polygon index
    RP = P[VP, 3] # Extract the rates of error by Voronoi polygon index
    # Create lists to hold interpolated values
    zc_i = []
    ec_i = []
    if extras == True:
        nnd = []
        rest = []
    # For all raster domain cells calculate value and rate of error estimates
    for i in range(c_xy.shape[0]):
        Dc_iC = distanceFromCell(c_xy[i,:], c_xy)
        Vpi = interpolationCellVoronoi(DPC, Dc_iC)
        zc_i.append(np.nanmean(ZP * Vpi))
        d_j = distanceFromCell(c_xy[i,:], P)
        VPd = d_j[VP]
        ec_i.append(np.nanmean(VPd * RP * Vpi))
        if extras == True:
            nnd.append(np.nanmean(VPd * Vpi))
            rest.append(np.nanmean(P[:,3][VP] * Vpi))
    # Reshape the sets of values to rasters
    nni = np.reshape(np.array(zc_i), (nRow, nCol))
    cvedf = np.reshape(np.array(ec_i), (nRow, nCol))
    if extras == True:    
        rest = np.reshape(np.array(rest), (nRow, nCol))
        nnd = np.reshape(np.array(nnd), (nRow, nCol))
        return((nni, cvedf, nnd, rest))
    else:
        return((nni, cvedf))
    
#------------------------------------------------------------------------------
