#------------------------------------------------------------------------------

# MIT License

# Copyright (c) 2020 Landcare Research New Zealand Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

# Python 3.5.5 | packaged by conda-forge | (default, Jul 24 2018, 01:52:17) [MSC v.1900 64 bit (AMD64)] on win32
import sys
sys.path.insert(0, "../")
import discreteNaturalNeighbourInterpolation as dnni
from nlmpy import nlmpy # version 0.1.5
import numpy as np # version 1.14.2
from scipy import spatial # version 1.1.0
import matplotlib as mpl # version 2.0.2
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.patches as patches

#------------------------------------------------------------------------------
    
# Define function to extract array values
def extractArrayValues(points, array, extent):
    # Extract the extent values
    xMin, xMax, yMin, yMax = extent
    n = points.shape[0]
    nRow, nCol = array.shape
    cellSize = (xMax - xMin) / float(nCol)
    values = []
    for i in range(n):
        row = nRow - int((float(points[i,1]) - yMin)/ cellSize) - 1
        col = int((float(points[i,0]) - xMin) / cellSize)
        values.append(array[row, col])
    return(np.array(values))

#------------------------------------------------------------------------------

# Specify the raster metadata
xMin = 10
yMin = 10
nRow = 120
nCol = 100
cellSize = 0.1
raster = [xMin, yMin, nRow, nCol, cellSize]
xMax = xMin + (nCol * cellSize)
yMax = yMin + (nRow * cellSize)
extent = [xMin, xMax, yMin, yMax]

#------------------------------------------------------------------------------

np.random.seed(4)

#------------------------------------------------------------------------------

# Specify some plotting stuff
markerSize = 4
# Check if LaTeX is installed to plot with consistent math font
from distutils.spawn import find_executable
if find_executable('latex'):
    mpl.pyplot.rc('text', usetex=True)
# If a LaTeX error is raised, the preceding 3 lines of code can be removed to
# reproduce figures with slightly different fonts or you could try adding
# mpl.pyplot.rcParams['text.usetex'] = False if on Matplotlib version >=3.x
mpl.pyplot.rc('font', family='serif', size=8)

#------------------------------------------------------------------------------

fig, axes = mpl.pyplot.subplots(nrows=2, ncols=5, figsize=(6, 4)) # figsize=(width, height) in inches

captions = ["(a)", "(b)", "(c)", "(d)", "(e)",
            "(f)", "(g)", "(h)", "(i)", "(j)"]

a = -1

for row in axes:
    
    np.random.seed(4)
    
    #--------------------------------------------------------------------------
    
    # Create the neutral landscape model
    a = a + 1
    if a == 0:
        h = 2
    else:
        h = 1
    nlm = nlmpy.mpd(nRow, nCol, h)
    # Create random points and extract the values
    if a == 0:
        n = 20
    else:
        n = 10
    x = np.random.uniform(xMin, xMax, n)
    y = np.random.uniform(yMin, yMax, n)
    xy = np.column_stack((x, y))
    z = extractArrayValues(xy, nlm, extent)
    xyz = np.column_stack((xy, z))
    # Calculate a convex hull for the data points
    convexhull = spatial.ConvexHull(xyz[:,:2])    
    # Do the natural neighbour interpolation
    nni, cvedf = dnni.nni(xyz, raster)
    # Calculate value error and error of errors
    valueError = np.abs(nni - nlm)
    errorOfErrors = np.abs(cvedf - valueError)

    #--------------------------------------------------------------------------

    # Plot NLM
    row[0].set_title(captions[a])
    im = row[0].imshow(nlm, interpolation='none', extent=extent, vmin = 0, vmax = 1,
               cmap=mpl.cm.YlOrRd)
    row[0].xaxis.set_ticks([])
    row[0].yaxis.set_ticks([])
    divider = make_axes_locatable(row[0])
    cax = divider.append_axes("bottom", size="4%", pad=0.1)
    cbar = fig.colorbar(im, ax=row[0], orientation='horizontal', cax=cax, ticks=[0.0, 0.5, 1.0])
    cbar.set_label("Actual value")
    # Define a function to point the z values of the data cells
    def plotValues(ax):
        for n in range(xyz.shape[0]):
            ax.plot(xyz[n,0], xyz[n,1], 'o', c='w', ms=markerSize,
                            markeredgewidth=0.5, markeredgecolor='k')
    plotValues(row[0])
    
    # Plot NNI
    a = a + 1
    row[1].set_title(captions[a])
    im = row[1].imshow(nni, interpolation='none', extent=extent, vmin = 0, vmax = 1,
                    cmap=mpl.cm.YlOrRd)
    row[1].xaxis.set_ticks([])
    row[1].yaxis.set_ticks([])
    divider = make_axes_locatable(row[1])
    cax = divider.append_axes("bottom", size="4%", pad=0.1)
    cbar = fig.colorbar(im, ax=row[1], orientation='horizontal', cax=cax, ticks=[0.0, 0.5, 1.0])
    cbar.set_label("Estimated value")
    plotValues(row[1])
    
    # Plot value error
    a = a + 1
    row[2].set_title(captions[a])
    im = row[2].imshow(valueError, interpolation='none', extent=extent, vmin = 0, vmax = 0.2,
                    cmap=mpl.cm.YlGn)
    row[2].xaxis.set_ticks([])
    row[2].yaxis.set_ticks([])
    divider = make_axes_locatable(row[2])
    cax = divider.append_axes("bottom", size="4%", pad=0.1)
    cbar = fig.colorbar(im, ax=row[2], orientation='horizontal', cax=cax, ticks=[0.0, 0.1, 0.2])
    cbar.set_ticklabels(['0.0', '0.1', r'$\geq 0.2$'])
    cbar.set_label("Value error")
    plotValues(row[2])
    row[2].add_patch(patches.Polygon(xyz[:,:2][convexhull.vertices], fc='none', ec='k'))
    
    # Plot estimated error
    a = a + 1
    row[3].set_title(captions[a])
    im = row[3].imshow(cvedf, interpolation='none', extent=extent, vmin = 0, vmax = 0.2,
                    cmap=mpl.cm.YlGn)
    row[3].xaxis.set_ticks([])
    row[3].yaxis.set_ticks([])
    divider = make_axes_locatable(row[3])
    cax = divider.append_axes("bottom", size="4%", pad=0.1)
    cbar = fig.colorbar(im, ax=row[3], orientation='horizontal', cax=cax, ticks=[0.0, 0.1, 0.2])
    cbar.set_ticklabels(['0.0', '0.1', r'$\geq 0.2$'])
    cbar.set_label("Estimated error")
    plotValues(row[3])
    
    # Plot error of errors
    a = a + 1
    row[4].set_title(captions[a])
    im = row[4].imshow(errorOfErrors, interpolation='none', extent=extent, vmin = 0, vmax = 0.2,
                    cmap=mpl.cm.BuPu)
    row[4].xaxis.set_ticks([])
    row[4].yaxis.set_ticks([])
    divider = make_axes_locatable(row[4])
    cax = divider.append_axes("bottom", size="4%", pad=0.1)
    cbar = fig.colorbar(im, ax=row[4], orientation='horizontal', cax=cax, ticks=[0.0, 0.1, 0.2])
    cbar.set_ticklabels(['0.0', '0.1', r'$\geq 0.2$'])
    cbar.set_label("Error of errors")
    plotValues(row[4])
    row[4].add_patch(patches.Polygon(xyz[:,:2][convexhull.vertices], fc='none', ec='k'))

#------------------------------------------------------------------------------

# Export figure
mpl.pyplot.tight_layout(pad=0.5)
mpl.pyplot.savefig('figure-5.png', dpi=300)

#------------------------------------------------------------------------------
