#------------------------------------------------------------------------------

# MIT License

# Copyright (c) 2020 Landcare Research New Zealand Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

# Python 3.5.5 | packaged by conda-forge | (default, Jul 24 2018, 01:52:17) [MSC v.1900 64 bit (AMD64)] on win32
import sys
sys.path.insert(0, "../")
import discreteNaturalNeighbourInterpolation as dnni
from nlmpy import nlmpy # version 0.1.5
import numpy as np # version 1.14.2
from scipy import spatial, stats # version 1.1.0
import matplotlib as mpl # version 2.0.2
from mpl_toolkits.axes_grid1 import make_axes_locatable

#------------------------------------------------------------------------------
    
# Define function to extract array values
def extractArrayValues(points, array, extent):
    # Extract the extent values
    xMin, xMax, yMin, yMax = extent
    n = points.shape[0]
    nRow, nCol = array.shape
    cellSize = (xMax - xMin) / float(nCol)
    values = []
    for i in range(n):
        row = nRow - int((float(points[i,1]) - yMin)/ cellSize) - 1
        col = int((float(points[i,0]) - xMin) / cellSize)
        values.append(array[row, col])
    return(np.array(values))
   
#------------------------------------------------------------------------------
    
# Specify the raster metadata
xMin = 10
yMin = 10
nRow = 120
nCol = 100
cellSize = 0.1
# Create a raster domain list
C = [xMin, yMin, nRow, nCol, cellSize]
xMax = xMin + (nCol * cellSize)
yMax = yMin + (nRow * cellSize)
extent = [xMin, xMax, yMin, yMax]
# Get xy coordinates for raster cells
c_xy = dnni.rasterDomainCellCoordinates(C)

#------------------------------------------------------------------------------

# Define number of experiments
experiments = 500
np.random.seed(1) # set random seed for reproducible results

#------------------------------------------------------------------------------

# Only process if the results file is missing
import os.path
if os.path.isfile("results.txt") == False:
    
    results = []
    
    for experiment in range(experiments):
        print(experiment)
        
        # Set the spatial autocorrelation and sample size
        h = np.random.uniform(low=0, high=2, size=1)[0]
        n = np.random.choice(np.arange(10,100), size=1)[0]
    
        # Create NLM and points            
        nlm = nlmpy.mpd(nRow, nCol, h)
        x = np.random.uniform(xMin, xMax, n)
        y = np.random.uniform(yMin, yMax, n)
        xy = np.column_stack((x, y))
        z = extractArrayValues(xy, nlm, extent)
        xyz = np.column_stack((xy, z))
        
        #----------------------------------------------------------------------
        
        # Get the cross-validated mean absolute error
        P = dnni.xyzToDataCells(xyz, C)
        cvmae = dnni.crossValidateMeanAbsoluteError(P, c_xy)
        
        #----------------------------------------------------------------------
        
        # Do the natural neighbour interpolation
        nni, cvedf = dnni.nni(xyz, C)
        # Calculate value error and error of errors
        valueError = np.abs(nni - nlm)
        errorOfErrors = np.abs(cvedf - valueError)
        
        #----------------------------------------------------------------------
        
        # Create convex hull
        P = dnni.xyzToDataCells(xyz, C)
        dt = spatial.Delaunay(P[:,:2])
        gridSpaceSimplex = dt.find_simplex(c_xy)
        ch = np.reshape(gridSpaceSimplex, (nRow, nCol)) * 1.0
        np.place(ch, ch>=0, 1.0)
        np.place(ch, ch==-1, np.nan)
        # Create non-convex hull
        nonch = np.ones((nRow, nCol))
        np.place(nonch, ch==1, np.nan)
        
        #----------------------------------------------------------------------

        # Mean absolute errors for value and error estimates
        nniMAEin = np.nanmean(np.abs(valueError * ch))
        errMAEin = np.nanmean(np.abs(errorOfErrors * ch))
        nniMAEout = np.nanmean(np.abs(valueError * nonch))
        errMAEout = np.nanmean(np.abs(errorOfErrors * nonch))
        results.append([n, h, cvmae, nniMAEin, errMAEin, nniMAEout, errMAEout])
        
    #--------------------------------------------------------------------------
                
    # Write out results                
    df = np.array(results)
    np.savetxt('results.txt', df, delimiter='\t')
    n = df[:,0] # Number of sample locations
    h = df[:,1] # The auto-correlation of the NLM
    cvmae = df[:,2] # The cross-validation mean absolute error
    nniMAEin = df[:,3] # The mean absolute error of predicted values in the convex hull
    errMAEin = df[:,4] # The mean absolute error of predicted errors in the convex hull
    nniMAEout = df[:,5] # The mean absolute error of predicted values outside the convex hull
    errMAEout = df[:,6] # The mean absolute error of predicted errors outside the convex hull
     
    #--------------------------------------------------------------------------

# Else read in the preprocessed results
else:
    df = np.loadtxt('results.txt', delimiter='\t')
    n = df[:,0] # Number of sample locations
    h = df[:,1] # The auto-correlation of the NLM
    cvmae = df[:,2] # The cross-validation mean absolute error
    nniMAEin = df[:,3] # The mean absolute error of predicted values in the convex hull
    errMAEin = df[:,4] # The mean absolute error of predicted errors in the convex hull
    nniMAEout = df[:,5] # The mean absolute error of predicted values outside the convex hull
    errMAEout = df[:,6] # The mean absolute error of predicted errors outside the convex hull
    
#------------------------------------------------------------------------------
# Plot figures
#------------------------------------------------------------------------------

# Check if LaTeX is installed to plot with consistent math font
from distutils.spawn import find_executable
if find_executable('latex'):
    mpl.pyplot.rc('text', usetex=True)
# If a LaTeX error is raised, the preceding 3 lines of code can be removed to
# reproduce figures with slightly different fonts or you could try adding
# mpl.pyplot.rcParams['text.usetex'] = False if on Matplotlib version >=3.x
mpl.pyplot.rc('font', family='serif', size=8)

#------------------------------------------------------------------------------
# Figure 6
#------------------------------------------------------------------------------

fig, axs = mpl.pyplot.subplots(figsize=(6, 7)) # figsize=(width, height) in inches
gs = mpl.gridspec.GridSpec(3, 2, height_ratios = [4,3,3])
cm = mpl.pyplot.cm.get_cmap('plasma')

ax = mpl.pyplot.subplot(gs[0, 0])
ax.set_title("(a)")
sc = ax.scatter(h, n, c=nniMAEin, vmin=0, vmax=0.15, s=20, cmap=cm)
ax.set_xlabel(r"$h$")
ax.set_ylabel(r"$n$")
ax.yaxis.set_ticks([10, 40, 70, 100])
divider = make_axes_locatable(ax)
cax = divider.append_axes("bottom", size="4%", pad=0.5)
cbar = fig.colorbar(sc, ax=ax, orientation='horizontal', cax=cax, ticks=[0.0, 0.05, 0.1, 0.15])
cbar.set_label("MAE of " + r"$e(\hat{z})$" + " inside convex hull")

ax = mpl.pyplot.subplot(gs[0, 1])
ax.set_title("(b)")
sc = ax.scatter(h, n, c=errMAEin, vmin=0, vmax=0.15, s=20, cmap=cm)
ax.set_xlabel(r"$h$")
ax.set_ylabel(r"$n$")
ax.yaxis.set_ticks([10, 40, 70, 100])
ax.yaxis.set_ticks([10, 40, 70, 100])
divider = make_axes_locatable(ax)
cax = divider.append_axes("bottom", size="4%", pad=0.5)
cbar = fig.colorbar(sc, ax=ax, orientation='horizontal', cax=cax, ticks=[0.0, 0.05, 0.1, 0.15])
cbar.set_label("MAE of " + r"$e(\hat{e})$" + " inside convex hull")

ax = mpl.pyplot.subplot(gs[1, 0])
ax.set_title("(c)")
sc = ax.scatter(nniMAEin, errMAEin, c='k', s=1)
ax.set_xlabel("MAE of " + r"$e(\hat{z})$" + " inside convex hull")
ax.set_ylabel("MAE of " + r"$e(\hat{e})$" + " inside convex hull")
ax.set_xlim([0, 0.15])
ax.set_ylim([0, 0.15])
ax.xaxis.set_ticks([0.0, 0.05, 0.1, 0.15])
ax.yaxis.set_ticks([0.0, 0.05, 0.1, 0.15])
gradient, intercept, r_value, p_value, std_err = stats.linregress(nniMAEin, errMAEin)
xMinMax = np.array([0, 0.15])
yMinMax = gradient * xMinMax + intercept
mpl.pyplot.plot(xMinMax, yMinMax, 'red', linewidth=1)
mpl.pyplot.text(0.12, 0.11, s = r"$r = $" + "%.2f" % r_value, color='r')

ax = mpl.pyplot.subplot(gs[1, 1])
ax.set_title("(d)")
sc = ax.scatter(nniMAEin, cvmae, c='k', s=1)
ax.set_xlabel("MAE of " + r"$e(\hat{z})$" + " inside convex hull")
ax.set_ylabel("Cross-validation MAE")
ax.set_xlim([0, 0.15])
ax.set_ylim([0, 0.21])
ax.xaxis.set_ticks([0.0, 0.05, 0.1, 0.15])
ax.yaxis.set_ticks([0.0, 0.07, 0.14, 0.21])
gradient, intercept, r_value, p_value, std_err = stats.linregress(nniMAEin, cvmae)
xMinMax = np.array([0, 0.15])
yMinMax = gradient * xMinMax + intercept
mpl.pyplot.plot(xMinMax, yMinMax, 'red', linewidth=1)
mpl.pyplot.text(0.12, 0.16, s = r"$r = $" + "%.2f" % r_value, color='red')
mpl.pyplot.plot([0, 1], [0, 1], 'grey', linewidth=1)
mpl.pyplot.text(0.13, 0.11, s = r"$x = y$", color='grey')

ax = mpl.pyplot.subplot(gs[2, 0])
ax.set_title("(e)")
sc = ax.scatter(nniMAEin, nniMAEout, c='k', s=1)
ax.set_xlabel("MAE of " + r"$e(\hat{z})$" + " inside convex hull")
ax.set_ylabel("MAE of " + r"$e(\hat{z})$" + " outside convex hull")
ax.set_xlim([0, 0.15])
ax.set_ylim([0, 0.27])
ax.xaxis.set_ticks([0.0, 0.05, 0.1, 0.15])
ax.yaxis.set_ticks([0.0, 0.09, 0.18, 0.27])
gradient, intercept, r_value, p_value, std_err = stats.linregress(nniMAEin, nniMAEout)
xMinMax = np.array([0, 0.15])
yMinMax = gradient * xMinMax + intercept
mpl.pyplot.plot(xMinMax, yMinMax, 'red', linewidth=1)
mpl.pyplot.text(0.12, 0.18, s = r"$r = $" + "%.2f" % r_value, color='red')
mpl.pyplot.plot([0, 1], [0, 1], 'grey', linewidth=1)
mpl.pyplot.text(0.13, 0.11, s = r"$x = y$", color='grey')

ax = mpl.pyplot.subplot(gs[2, 1])
ax.set_title("(f)")
sc = ax.scatter(errMAEin, errMAEout, c='k', s=1)
ax.set_xlabel("MAE of " + r"$e(\hat{e})$" + " inside convex hull")
ax.set_ylabel("MAE of " + r"$e(\hat{e})$" + " outside convex hull")
ax.set_xlim([0, 0.15])
ax.set_ylim([0, 0.27])
ax.xaxis.set_ticks([0.0, 0.05, 0.1, 0.15])
ax.yaxis.set_ticks([0.0, 0.09, 0.18, 0.27])
gradient, intercept, r_value, p_value, std_err = stats.linregress(errMAEin, errMAEout)
xMinMax = np.array([0, 0.15])
yMinMax = gradient * xMinMax + intercept
mpl.pyplot.plot(xMinMax, yMinMax, 'red', linewidth=1)
mpl.pyplot.text(0.12, 0.20, s = r"$r = $" + "%.2f" % r_value, color='red')
mpl.pyplot.plot([0, 1], [0, 1], 'grey', linewidth=1)
mpl.pyplot.text(0.13, 0.11, s = r"$x = y$", color='grey')

mpl.pyplot.tight_layout(pad=0.5)
mpl.pyplot.savefig('figure-6.png', dpi=300)

#------------------------------------------------------------------------------
