#------------------------------------------------------------------------------

# MIT License

# Copyright (c) 2020 Landcare Research New Zealand Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#------------------------------------------------------------------------------

# Python 3.5.5 | packaged by conda-forge | (default, Jul 24 2018, 01:52:17) [MSC v.1900 64 bit (AMD64)] on win32
import sys
sys.path.insert(0, "../")
import discreteNaturalNeighbourInterpolation as dnni
import numpy as np # version 1.14.2
import matplotlib as mpl # version 2.0.2
from mpl_toolkits.axes_grid1 import make_axes_locatable

#------------------------------------------------------------------------------

# Specify the raster metadata
xMin = 10
yMin = 10
nRow = 120
nCol = 100
cellSize = 0.1
# Create a raster domain list
C = [xMin, yMin, nRow, nCol, cellSize]
# Get xy coordinates for raster cells
c_xy = dnni.rasterDomainCellCoordinates(C)
# Import point points
xyz = np.loadtxt("points.txt", delimiter=",", skiprows=1)
# Convert locations to data cell centres
P = dnni.xyzToDataCells(xyz, C)
n = P.shape[0]
# Do cross-validation to add rates of errors to P
P = dnni.crossValidateRateOfError(P, c_xy)
# Calculate the Voronoi diagram (VP), and the distance from the data cells to
# the raster domain cells (DPC)
VP, DPC = dnni.discreteVoronoiDiagram(P, c_xy)
ZP = P[VP, 2] # Extract values by Voronoi polygon index
# Specify interpolation cell
c_i = np.array([15.05, 16.05])
# Calculate the distances from the interplation cell (c_i) to the raster 
# domain cells (C)
Dc_iC = dnni.distanceFromCell(c_i, c_xy)
# Determine the interpolation cell discrete Voronoi polygon
Vc_i = dnni.interpolationCellVoronoi(DPC, Dc_iC)
# Calculate distances from interpolation cell to data cells
d_j = dnni.distanceFromCell(c_i, P)
VPd = d_j[VP] # Extract distances by Voronoi polygon index
# Calculate the:
# - natural natural neighbour interpolation (nni)
# - cross-validaton error-distance field (cvedf)
# - natural neighbour distance (nnd)
# - estimated rate of absolute error (rest)
nni, cvedf, nnd, rest = dnni.nni(xyz, C, extras = True)

#------------------------------------------------------------------------------
# Plot figures
#------------------------------------------------------------------------------

# Specify some plotting stuff
extent = [xMin, xMin + (nCol * cellSize), yMin, yMin + (nRow * cellSize)]
colours = ["#1b9e77" ,"#d95f02", "#7570b3", "#e7298a",
           "#66a61e", "#e6ab02", "#a6761d", "#666666"]
cmapClas = mpl.colors.ListedColormap(colours[:n])
bounds=[0,1,2,3,4,5,6,7,8]
norm = mpl.colors.BoundaryNorm(bounds, n)
markerSize = 4
# Check if LaTeX is installed to plot with consistent math font
from distutils.spawn import find_executable
if find_executable('latex'):
    mpl.pyplot.rc('text', usetex=True)
# If a LaTeX error is raised, the preceding 3 lines of code can be removed to
# reproduce figures with slightly different fonts or you could try adding
# mpl.pyplot.rcParams['text.usetex'] = False if on Matplotlib version >=3.x
mpl.pyplot.rc('font', family='serif', size=8)

#------------------------------------------------------------------------------
# Plot discrete natural neighbour interpolation method
#------------------------------------------------------------------------------

fig, ((ax1, ax2, ax3), (ax4, ax5, ax6)) = mpl.pyplot.subplots(nrows=2, ncols=3, 
                                                              figsize=(6, 6)) # figsize=(width, height) in inches

# Plot VP
ax1.set_title("(a)")
im = ax1.imshow(np.reshape(VP, (nRow, nCol)), 
                interpolation='none', extent=extent, cmap=cmapClas, norm=norm)
ax1.xaxis.set_ticks([10,15,20])
ax1.yaxis.set_ticks([10,16,22])
ax1.xaxis.set_ticklabels([])
for p in range(n):
    ax1.plot(P[p,0], P[p,1], 's', c=colours[p], ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')
    label = r"$p_" + str(p+1) + "$"
    ax1.text(P[p,0], P[p,1] + 0.5, label, fontsize=6,
             horizontalalignment='center', verticalalignment='center')
divider = make_axes_locatable(ax1)
cax = divider.append_axes("bottom", size="2.5%", pad=0.1)
cbar = fig.colorbar(im, ax=ax1, orientation='horizontal', cax=cax,
                    ticks=np.arange(np.min(VP),np.max(VP)+2)+0.5)
cbar.ax.set_xticklabels([1,2,3,4,5,6,7,8])
cbar.set_label("Index " + r"$(i)$")

# Plot DP
ax2.set_title("(b)")
maxDist = np.ceil(np.max(DPC))
im = ax2.imshow(np.reshape(DPC, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = maxDist,
                cmap=mpl.cm.YlGnBu)
ax2.xaxis.set_ticks([10,15,20])
ax2.yaxis.set_ticks([10,16,22])
ax2.xaxis.set_ticklabels([])
ax2.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax2)
cax = divider.append_axes("bottom", size="2.5%", pad=0.1)
cbar = fig.colorbar(im, ax=ax2, orientation='horizontal', cax=cax)
cbar.set_label("Euclidean distance " + r"$(d)$")
for p in range(n):
    ax2.plot(P[p,0], P[p,1], 's', c=mpl.cm.YlGnBu(0), ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')

# Plot VPz
ax3.set_title("(c)")
im = ax3.imshow(np.reshape(ZP, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = 1,
                cmap=mpl.cm.YlOrRd)
ax3.xaxis.set_ticks([10,15,20])
ax3.yaxis.set_ticks([10,16,22])
ax3.xaxis.set_ticklabels([])
ax3.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax3)
cax = divider.append_axes("bottom", size="2.5%", pad=0.1)
cbar = fig.colorbar(im, ax=ax3, orientation='horizontal', cax=cax)
cbar.set_label("Value " + r"$(z)$")
c = mpl.cm.YlOrRd(P[:,2])

# Define a function to point the z values of the data cells
def plotValues(ax, labelPoint=True):
    for p in range(n):
        ax.plot(P[p,0], P[p,1], 's', c=c[p], ms=markerSize,
                 markeredgewidth=0.5, markeredgecolor='k')
        if labelPoint == True:
            label = r"$z_" + str(p+1) + "=" + str(np.round(P[p,2], 2)) + "$"
            ax.text(P[p,0], P[p,1] + 0.5, label, fontsize=6,
                     horizontalalignment='center', verticalalignment='center')
plotValues(ax3)

# Plot Dc_i
ax4.set_title("(d)")
im = ax4.imshow(np.reshape(Dc_iC, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = maxDist,
                cmap=mpl.cm.YlGnBu)
ax4.xaxis.set_ticks([10,15,20])
ax4.yaxis.set_ticks([10,16,22])
divider = make_axes_locatable(ax4)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax4, orientation='horizontal', cax=cax,
                    ticks=[0,1,2,3,4,5], )
cbar.set_ticklabels(['0', '1', '2', '3', '4', r'$\geq 5$'])
cbar.set_label("Euclidean distance " + r"$(d)$")
ax4.plot(c_i[0], c_i[1], marker='s', c='w', ms=markerSize,
         markeredgewidth=0.5, markeredgecolor='k')
ax4.text(c_i[0], c_i[1] + 0.5, r'$c_i$', 
         horizontalalignment='center', verticalalignment='center')


# Plot Vc_z
ax5.set_title("(e)")
im = ax5.imshow(np.reshape(ZP * Vc_i, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = 1,
                cmap=mpl.cm.YlOrRd)
ax5.xaxis.set_ticks([10,15,20])
ax5.yaxis.set_ticks([10,16,22])
ax5.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax5)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax5, orientation='horizontal', cax=cax)
cbar.set_label("Value " + r"$(z)$")
ax5.plot(c_i[0], c_i[1], marker='s', c='w', ms=markerSize,
         markeredgewidth=0.5, markeredgecolor='k')
ax5.text(c_i[0], c_i[1] + 0.5, r'$c_i$', 
         horizontalalignment='center', verticalalignment='center')

# Plot nni
ax6.set_title("(f)")
im = ax6.imshow(nni, 
                interpolation='none', extent=extent, vmin = 0, vmax = 1,
                cmap=mpl.cm.YlOrRd)
ax6.xaxis.set_ticks([10,15,20])
ax6.yaxis.set_ticks([10,16,22])
ax6.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax6)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax6, orientation='horizontal', cax=cax)
cbar.set_label("Estimated value " + r"$(\hat{z})$")
c = mpl.cm.YlOrRd(P[:,2])
plotValues(ax6, labelPoint=False)

mpl.pyplot.tight_layout(pad=0.5)
mpl.pyplot.savefig('figure-1.png', dpi=300)

#------------------------------------------------------------------------------
# Plot natural neighbour distance method
#------------------------------------------------------------------------------

fig, ((ax1, ax2, ax3)) = mpl.pyplot.subplots(nrows=1, ncols=3, 
                                             figsize=(6, 3)) # figsize=(width, height) in inches

# Plot VP_d
ax1.set_title("(a)")
im = ax1.imshow(np.reshape(VPd, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = np.ceil(np.max(Dc_iC)),
                cmap=mpl.cm.YlGnBu)
ax1.xaxis.set_ticks([10,15,20])
ax1.yaxis.set_ticks([10,16,22])
divider = make_axes_locatable(ax1)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax1, orientation='horizontal', cax=cax)
cbar.set_label("Euclidean distance " + r"$(d)$")
c = mpl.cm.YlGnBu(d_j / np.ceil(np.max(Dc_iC)))
for p in range(n):
    ax1.plot(P[p,0], P[p,1], 's', c=c[p], ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')
    label = r"$d_" + str(p+1) + "=" + str(np.round(d_j[p], 1)) + "$"
    ax1.text(P[p,0], P[p,1] + 0.5, label, fontsize=6,
             horizontalalignment='center', verticalalignment='center')
    ax1.plot([c_i[0], P[p,0]], [c_i[1], P[p,1]], 
             linestyle='--', linewidth=0.5, color='k')
ax1.plot(c_i[0], c_i[1], marker='s', c='w', ms=markerSize,
         markeredgewidth=0.5, markeredgecolor='k')
ax1.text(c_i[0] - 0.5, c_i[1] - 0.1, r'$c_i$', 
         horizontalalignment='center', verticalalignment='center')

# Plot Vc_d
ax2.set_title("(b)")
im = ax2.imshow(np.reshape(VPd * Vc_i, (nRow, nCol)), 
                interpolation='none', extent=extent, vmin = 0, vmax = np.ceil(np.max(Dc_iC)),
           cmap=mpl.cm.YlGnBu)
ax2.xaxis.set_ticks([10,15,20])
ax2.yaxis.set_ticks([10,16,22])
ax2.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax2)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax2, orientation='horizontal', cax=cax)
cbar.set_label("Euclidean distance " + r"$(d)$")
ax2.plot(c_i[0], c_i[1], marker='s', c='w', ms=markerSize,
         markeredgewidth=0.5, markeredgecolor='k')
ax2.text(c_i[0], c_i[1] + 0.5, r'$c_i$', 
         horizontalalignment='center', verticalalignment='center')

# Plot DP
ax3.set_title("(c)")
im = ax3.imshow(nnd, 
                interpolation='none', extent=extent, vmin = 0, vmax = np.ceil(np.max(nnd)),
           cmap=mpl.cm.YlGnBu)
ax3.xaxis.set_ticks([10,15,20])
ax3.yaxis.set_ticks([10,16,22])
divider = make_axes_locatable(ax3)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax3, orientation='horizontal', cax=cax)
ax3.yaxis.set_ticklabels([])
cbar.set_label("Natural neighbour distance " + r"$(\delta)$")
for p in range(n):
    ax3.plot(P[p,0], P[p,1], 's', c=mpl.cm.YlGnBu(0), ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')

mpl.pyplot.tight_layout(pad=0.5)
mpl.pyplot.savefig('figure-2.png', dpi=300)

#------------------------------------------------------------------------------
# Plot natural neighbour error estimates
#------------------------------------------------------------------------------

fig, ((ax1, ax2)) = mpl.pyplot.subplots(nrows=1, ncols=2, 
                                             figsize=(4, 3)) # figsize=(width, height) in inches


# Plot interpolated rate of error
ax1.set_title("(a)")
im = ax1.imshow(rest, 
                interpolation='none', extent=extent, vmin=0, vmax=0.25,
                cmap=mpl.cm.YlGn)
ax1.xaxis.set_ticks([10,15,20])
ax1.yaxis.set_ticks([10,16,22])
divider = make_axes_locatable(ax1)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax1, orientation='horizontal', cax=cax)
cbar.set_label("Estimated rate of error " + r"$(\hat{r})$")
c = mpl.cm.YlGn(P[:,3] / 0.25)

for p in range(n):
    ax1.plot(P[p,0], P[p,1], 's', c=c[p], ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')
    label = r"$r_" + str(p+1) + "=" + str(np.round(P[p,3], 2)) + "$"
    ax1.text(P[p,0], P[p,1] + 0.5, label, fontsize=6,
             horizontalalignment='center', verticalalignment='center')

# Plot error-distance estimate
ax2.set_title("(b)")
im = ax2.imshow(cvedf, 
                interpolation='none', extent=extent, vmin=0, vmax=0.6,
                cmap=mpl.cm.YlGn)
ax2.xaxis.set_ticks([10,15,20])
ax2.yaxis.set_ticks([10,16,22])
ax2.yaxis.set_ticklabels([])
divider = make_axes_locatable(ax2)
cax = divider.append_axes("bottom", size="2.5%", pad=0.25)
cbar = fig.colorbar(im, ax=ax2, orientation='horizontal', cax=cax)
cbar.set_label("Error-distance field " + r"$(\hat{e})$")
for p in range(n):
    ax2.plot(P[p,0], P[p,1], 's', c=mpl.cm.YlGn(0), ms=markerSize,
             markeredgewidth=0.5, markeredgecolor='k')
    
mpl.pyplot.tight_layout(pad=0.5)
mpl.pyplot.savefig('figure-3.png', dpi=300)

#------------------------------------------------------------------------------
